from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel, SecretStr


class PrometheusApis(Enum):
    QUERY = 0
    QUERY_RANGE = 1
    LABELS = 2
    FLAGS = 3
    VM_FLAGS = 4


class PrometheusConfig(BaseModel):
    url: str
    disable_ssl: bool = False
    headers: Dict[str, str] = {}
    prometheus_auth: Optional[SecretStr] = None
    prometheus_url_query_string: Optional[str] = None
    additional_labels: Optional[Dict[str, str]] = None
    supported_apis: List[PrometheusApis] = [
        PrometheusApis.QUERY,
        PrometheusApis.QUERY_RANGE,
        PrometheusApis.LABELS,
        PrometheusApis.FLAGS,
    ]
    query_step: str = "5m"
    query_interval: str = "1d"


class AWSPrometheusConfig(PrometheusConfig):
    access_key: str
    secret_access_key: str
    service_name: str = "aps"
    aws_region: str
    supported_apis: List[PrometheusApis] = [
        PrometheusApis.QUERY,
        PrometheusApis.QUERY_RANGE,
        PrometheusApis.LABELS,
    ]


class CoralogixPrometheusConfig(PrometheusConfig):
    prometheus_token: str
    supported_apis: List[PrometheusApis] = [
        PrometheusApis.QUERY,
        PrometheusApis.QUERY_RANGE,
        PrometheusApis.LABELS,
    ]


class VictoriaMetricsPrometheusConfig(PrometheusConfig):
    supported_apis: List[PrometheusApis] = [
        PrometheusApis.QUERY,
        PrometheusApis.QUERY_RANGE,
        PrometheusApis.LABELS,
        PrometheusApis.VM_FLAGS,
    ]


# Does not support labels according to the docs, See below for apis
# https://learn.microsoft.com/en-us/azure/azure-monitor/essentials/prometheus-api-promql#supported-apis
class AzurePrometheusConfig(PrometheusConfig):
    azure_resource: str
    azure_metadata_endpoint: str
    azure_token_endpoint: str
    azure_use_managed_id: Optional[str] = None
    azure_client_id: Optional[str] = None
    azure_tenant_id: Optional[str] = None
    azure_client_secret: Optional[str] = None
    supported_apis: List[PrometheusApis] = [
        PrometheusApis.QUERY,
        PrometheusApis.QUERY_RANGE,
    ]



import logging
from typing import Dict, no_type_check

import requests

from prometrix.models.prometheus_config import (AzurePrometheusConfig,
                                                CoralogixPrometheusConfig,
                                                PrometheusConfig)


class PrometheusAuthorization:
    bearer_token: str = ""

    @classmethod
    def azure_authorization(cls, config: PrometheusConfig) -> bool:
        if not isinstance(config, AzurePrometheusConfig):
            return False
        return (config.azure_client_id != "" and config.azure_tenant_id != "") and (
            config.azure_client_secret != "" or config.azure_use_managed_id != ""
        )

    @classmethod
    def get_authorization_headers(cls, config: PrometheusConfig) -> Dict:
        if isinstance(config, CoralogixPrometheusConfig):
            return {"token": config.prometheus_token}
        elif config.prometheus_auth:
            return {"Authorization": config.prometheus_auth.get_secret_value()}
        elif cls.azure_authorization(config):
            return {"Authorization": (f"Bearer {cls.bearer_token}")}
        else:
            return {}

    @no_type_check
    @classmethod
    def _get_azure_metadata_endpoint(cls, config: PrometheusConfig):
        return requests.get(
            url=config.azure_metadata_endpoint,
            headers={
                "Metadata": "true",
            },
            data={
                "api-version": "2018-02-01",
                "client_id": config.azure_client_id,
                "resource": config.azure_resource,
            },
        )

    @no_type_check
    @classmethod
    def _post_azure_token_endpoint(cls, config: PrometheusConfig):
        return requests.post(
            url=config.azure_token_endpoint,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={
                "grant_type": "client_credentials",
                "client_id": config.azure_client_id,
                "client_secret": config.azure_client_secret,
                "resource": config.azure_resource,
            },
        )

    @classmethod
    def request_new_token(cls, config: PrometheusConfig) -> bool:
        if cls.azure_authorization(config) and isinstance(
            config, AzurePrometheusConfig
        ):
            try:
                if config.azure_use_managed_id:
                    res = cls._get_azure_metadata_endpoint(config)
                else:
                    res = cls._post_azure_token_endpoint(config)
            except Exception:
                logging.exception(
                    "Unexpected error when trying to generate azure access token."
                )
                return False

            if not res.ok:
                logging.error(f"Could not generate an azure access token. {res.reason}")
                return False

            cls.bearer_token = res.json().get("access_token")
            logging.info("Generated new azure access token.")
            return True

        return False



import json
from typing import Dict, List, Optional

PrometheusMetric = Dict[str, str]


class PrometheusScalarValue:
    def __init__(self, raw_scalar_list: List):
        """
        Initialize a Prometheus scalar value.
        The scalar is expected to be a list where the first element is a timestamp and the second is the value.
        """
        if len(raw_scalar_list) != 2:
            raise ValueError(f"Invalid prometheus scalar value {raw_scalar_list}")
        self.timestamp = float(raw_scalar_list[0])
        self.value = str(raw_scalar_list[1])

    def to_dict(self):
        """ Convert scalar value to a dictionary for JSON """
        return {
            "timestamp": self.timestamp,
            "value": self.value
        }

class PrometheusSeries:
    def __init__(self, metric: Dict[str, str], values: List):
        """
        Initialize a Prometheus series object.
        :param metric: Dictionary of metric labels.
        :param values: List of [timestamp, value] pairs.
        """
        self.metric = metric
        self.timestamps = [float(value[0]) for value in values]
        self.values = [str(value[1]) for value in values]

    def to_dict(self):
        """ Convert series object to a dictionary for JSON """
        return {
            "metric": self.metric,
            "timestamps": self.timestamps,
            "values": self.values
        }

class PrometheusQueryResult:
    def __init__(self, data: Dict):
        result = data.get("result", None)
        result_type = data.get("resultType", None)

        if not result_type:
            raise ValueError("resultType missing")
        if result is None:
            raise ValueError("result object missing")

        self.result_type = result_type
        self.vector_result = None
        self.series_list_result = None
        self.scalar_result = None
        self.string_result = None

        if result_type == "string" or result_type == "error":
            self.string_result = str(result)
        elif result_type == "scalar" and isinstance(result, list):
            self.scalar_result = PrometheusScalarValue(result).to_dict()
        elif result_type == "vector" and isinstance(result, list):
            self.vector_result = self._format_vector(result)
        elif result_type == "matrix" and isinstance(result, list):
            self.series_list_result = self._format_series(result)
        else:
            raise ValueError("result or returnType is invalid")

    def _format_vector(self, vector: List):
        """ Convert vector result into a list of dictionaries for JSON """
        return [
            {
                "metric": vector_item["metric"],
                "value": PrometheusScalarValue(vector_item["value"]).to_dict()
            }
            for vector_item in vector
        ]

    def _format_series(self, series: List):
        """ Convert matrix (series) result into a list of PrometheusSeries dictionaries for JSON """
        return [
            PrometheusSeries(series_item["metric"], series_item["values"]).to_dict()
            for series_item in series
        ]

    def to_json(self):
        """ Return the result as a JSON-compatible dictionary """
        return json.dumps({
            "result_type": self.result_type,
            "vector_result": self.vector_result,
            "series_list_result": self.series_list_result,
            "scalar_result": self.scalar_result,
            "string_result": self.string_result
        }, indent=4)


class MetricsNotFound(Exception):
    """
    An exception raised when Metrics service is not found.
    """

    pass


class PrometheusNotFound(MetricsNotFound):
    """
    An exception raised when Prometheus is not found.
    """

    pass


class VictoriaMetricsNotFound(MetricsNotFound):
    """
    An exception raised when Victoria Metrics is not found.
    """

    pass


class ThanosMetricsNotFound(MetricsNotFound):
    """
    An exception raised when Thanos is not found.
    """

    pass


class PrometheusFlagsConnectionError(Exception):
    """
    Exception, when Prometheus flag or AlertManager flag api cannot be reached
    """

    pass
